#include "stdafx.h"
#include "mesh.h"
#include "shadowmesh.h"

#include "shaderconstants.h"
#include "shadowVS.h"
#include "meshVS.h"
#include "ambientVS.h"
#include "shadowmeshVS.h"

class BuggyFollower
{
public:
	oVector3 buggyPosition, buggyVelocity;
	oVector2 buggyDirection;
	oVector3 followPosition, followVelocity;

	oReal followDistance, followHeight;

private:
	oReal omegaSq, beta;

public:
	BuggyFollower(const oReal& followDistance_, const oReal& followHeight_, const oReal& omegaSq_) 
	  : followDistance(followDistance_), followHeight(followHeight_), omegaSq(omegaSq_)
	{
		beta = _R(2) * oRealSqrt(omegaSq); // critical damping
	}

    void UpdatePosition(const oReal& timeStep, const bool& useDirection)
	{
		oVector3 a;

		// compute the position-dependent acceleration
		if(useDirection)
		{
			a.x = -omegaSq * (followPosition.x - (buggyPosition.x - followDistance * buggyDirection.x));
			a.z = -omegaSq * (followPosition.z - (buggyPosition.z - followDistance * buggyDirection.y));
			a.y = -omegaSq * (followPosition.y - (buggyPosition.y + followHeight));
		}
		else
		{
			oReal r = oRealSqrt(buggyVelocity.x * buggyVelocity.x + buggyVelocity.z * buggyVelocity.z);
			if(r > _R(2.5)) // > 2.5m/s
			{
				a.x = -omegaSq * (followPosition.x - (buggyPosition.x - followDistance * buggyVelocity.x / r));
				a.z = -omegaSq * (followPosition.z - (buggyPosition.z - followDistance * buggyVelocity.z / r));
				a.y = -omegaSq * (followPosition.y - (buggyPosition.y + followHeight));
			}
			else
			{
				oVector3 offset = followPosition - buggyPosition;
				r = oRealSqrt(offset.x * offset.x + offset.z * offset.z);
				if(r > _R(0.001)) // > 1mm
				{
					oReal accel = -omegaSq * (r - followDistance) / r;
					a.x = offset.x * accel;
					a.z = offset.z * accel;
				}
				a.y = -omegaSq * (offset.y - followHeight);
			}
		}

		// compute the velocity-dependent acceleration
		a -= beta * (followVelocity - buggyVelocity);

		// update by one timestep
		followVelocity += a * timeStep;
		followPosition += followVelocity * timeStep;
	}
};

// ODE globals
static SmartWorldID world;
static SmartSpaceID space;
static SmartJointGroupID contactGroup;

// ground and environment objects
static SmartGeomID groundGeom, wallGeom[5], envGeomGroup;
static Mesh groundMesh;

// ramp objects
static const oReal rampThickness = _R(0.4);
static const oReal supportThickness = _R(0.2);

static Mesh tableMeshes[4];
static ShadowMesh tableShadowMeshes[4];
static SmartGeomID tableGeom[18];

static Mesh jumpMeshes[2];
static ShadowMesh jumpShadowMeshes[2];
static SmartGeomID jumpGeom[12];

// car objects
static SmartGeomID carGeom[5], carGeomGroup;
static SmartBodyID carBodies[5];
static SmartJointID carJoints[4];
static Mesh carBodyMesh, carWheelMesh;
static ShadowMesh carBodyShadowMesh, carWheelShadowMesh;

// beachballs
static SmartGeomID ballGeom[27], ballGuardGeom[4];
static SmartBodyID ballBodies[27];
static Mesh ballMesh, ballGuardMesh;
static ShadowMesh ballShadowMesh, ballGuardShadowMesh;

// all static and dynamic meshes
static oStack<GeomMesh*, oOperatorDeleter<GeomMesh> > staticMeshes, dynamicMeshes;

// light and shadow volumes
static const ulong maxNumLights = 2;
static ulong numLights = 1;
static float lightColour[] = { _R(0.8), _R(0.8), _R(0.8), _R(1) };
static float ambientColour[] = { _R(0.4), _R(0.4), _R(0.4), _R(1) };
static const oVector3 baseLights[] = { 
	oVector3(_R(0.353553), _R(0.707107), _R(0.612372)),
	oVector3(_R(0.5), _R(0.707107), _R(-0.5)) 
};
static float volumeFilter[] = { 
	_R(1), _R(1), _R(0.5), _R(0.1), 
	_R(0.5), _R(1), _R(1), _R(0.1)
};
static uchar volumeKeys[] = { DIK_F1, DIK_F2 };
static oVector3 lights[maxNumLights];
static oQuaternion lightRotations[maxNumLights];
static oStack<GeomShadowVolume*, oOperatorDeleter<GeomShadowVolume> > staticShadows[maxNumLights], dynamicShadows[maxNumLights];

////////////////////////////////////////////////////////////////////////////
// Added by NateW
//
// Read the joystick X, Y, Z, and R positions, and scale them to +/- 1.0.
// Results are stored in the application object's m_ControlInput member.

#pragma warning(disable: 4201) // nonstandard extension used : nameless struct/union
#include <mmsystem.h>
#pragma warning(default: 4201)
typedef double real;

class ControlInput
{
public:
	real m_rJoystickX;
	real m_rJoystickY;
	real m_rJoystickZ;
	real m_rJoystickR;
};

static ControlInput g_ControlInput;

void RefreshControlInput ()
{
	JOYINFOEX state;
	MMRESULT mmr = 0;

	state.dwSize = sizeof (JOYINFOEX);
	state.dwFlags = JOY_RETURNALL;

	mmr = joyGetPosEx (JOYSTICKID1, &state);

#define GetData(x) \
	g_ControlInput.m_rJoystick##x = state.dw##x##pos; \
	g_ControlInput.m_rJoystick##x /= (0xFFFF / 2); \
	g_ControlInput.m_rJoystick##x -= 1;

	/* 
	 * Si added a bit of error handling here for those of us without joysticks.
	 * (the UI goes a bit nuts otherwise)
	 */
	if(mmr == JOYERR_NOERROR)
	{
		GetData(X);
		GetData(Y);
		GetData(Z);
		GetData(R);
	}
	else
		memset(&g_ControlInput, 0, sizeof(ControlInput));
}

// End of NateW contribution
////////////////////////////////////////////////////////////////////////////

void CreateWorld()
{
	static const oVector2 groundExtents(_R(50), _R(30));

	world = dWorldCreate();
	space = dHashSpaceCreate();
	contactGroup = dJointGroupCreate(0);		
	dWorldSetGravity(world, 0, -9.81, 0);

	// create the ground and (invisible) walls
	groundGeom = dCreatePlane(0,  0, 1,  0, 0); // x,y,z,d
	wallGeom[0] = dCreatePlane(0,  1, 0,  0, -groundExtents.x); // x,y,z,d
	wallGeom[1] = dCreatePlane(0, -1, 0,  0, -groundExtents.x); // x,y,z,d
	wallGeom[2] = dCreatePlane(0,  0, 0,  1, -groundExtents.y); // x,y,z,d
	wallGeom[3] = dCreatePlane(0,  0, 0, -1, -groundExtents.y); // x,y,z,d
	wallGeom[4] = dCreatePlane(0,  0, -1,  0, -40); // x,y,z,d = 40m cieling

	envGeomGroup = dCreateGeomGroup(space);
	dGeomGroupAdd(envGeomGroup, groundGeom);
    for(ulong i = 0; i < 5; ++i)
		dGeomGroupAdd(envGeomGroup, wallGeom[i]);

	// create the ground mesh
	groundMesh.CreateGroundPlane(groundExtents);
	staticMeshes.Push(new GeomMesh(groundGeom, &groundMesh));
}

void CreateBuggy()
{
	static const oVector3 carScale(_R(2.4), _R(0.4), _R(1.4));
	static const oReal wheelRadius = _R(0.3), wheelWidth = _R(0.25);
	static const wheelSegments = 24;

	// car body
	dMass m;
	dMassSetBox(&m, 1, carScale.x, carScale.y, carScale.z); // density,lx,ly,lz
	dMassAdjust(&m, 50); // mass

	carBodies[0] = dBodyCreate(world);
	dBodySetMass(carBodies[0], &m);
	dBodySetPosition(carBodies[0], 0, 3, 10); // x,y,z

	carGeom[0] = dCreateBox(0, carScale.x, carScale.y, carScale.z); // lx,ly,lz
	dGeomSetBody(carGeom[0], carBodies[0]);

	// wheel bodies
	dMassSetSphere(&m, 1, wheelRadius); // density, radius
	dMassAdjust(&m, 10); // mass
	dQuaternion q;
	dQFromAxisAndAngle(q, 1, 0, 0, M_PI * 0.5);
	for(ulong i = 1; i <= 4; ++i)
	{
		carBodies[i] = dBodyCreate(world);
		dBodySetMass(carBodies[i], &m);
		dBodySetQuaternion(carBodies[i], q);
		carGeom[i] = dCreateSphere(0, wheelRadius);
		dGeomSetBody(carGeom[i], carBodies[i]);
	}
	dBodySetPosition(carBodies[1], -0.9, 2.6,  10.7); // x,y,z
	dBodySetPosition(carBodies[2], -0.9, 2.6,   9.3); // x,y,z
	dBodySetPosition(carBodies[3],  0.9, 2.6,  10.7); // x,y,z
	dBodySetPosition(carBodies[4],  0.9, 2.6,   9.3); // x,y,z

	// wheel joints
	for(ulong i = 0; i < 4; ++i)
	{
		carJoints[i] = dJointCreateHinge2(world, 0);
		dJointAttach(carJoints[i], carBodies[0], carBodies[i+1]);
		const dReal* const wPos = dBodyGetPosition(carBodies[i+1]);
		dJointSetHinge2Anchor(carJoints[i], wPos[0], wPos[1], wPos[2]);
		dJointSetHinge2Axis1(carJoints[i], 0, 1, 0);
		dJointSetHinge2Axis2(carJoints[i], 0, 0, ((i % 2) == 0) ? -1 : 1);

		dJointSetHinge2Param(carJoints[i], dParamLoStop, 0);
		dJointSetHinge2Param(carJoints[i], dParamHiStop, 0);
		dJointSetHinge2Param(carJoints[i], dParamFMax, 50);

		dJointSetHinge2Param(carJoints[i], dParamVel2, 0);
		dJointSetHinge2Param(carJoints[i], dParamFMax2, 80);

		dJointSetHinge2Param(carJoints[i], dParamSuspensionERP, 0.25);
		dJointSetHinge2Param(carJoints[i], dParamSuspensionCFM, 0.004);
	}

	// geometry goup for the car
	carGeomGroup = dCreateGeomGroup(space);  
	for(ulong i = 0; i < 5; ++i)
		dGeomGroupAdd(carGeomGroup, carGeom[i]);

	// car meshes
	carBodyMesh.CreateBox(carScale);
	carWheelMesh.CreateCylinder(wheelRadius, wheelWidth, wheelSegments);
	dynamicMeshes.Push(new GeomMesh(carGeom[0], &carBodyMesh));
	for(ulong i = 1; i < 5; ++i)
		dynamicMeshes.Push(new GeomMesh(carGeom[i], &carWheelMesh));

	// car shadows
    carBodyShadowMesh.CreateBox(carScale);
	carWheelShadowMesh.CreateCylinder(wheelRadius, wheelWidth, wheelSegments);
	for(ulong lightIndex = 0; lightIndex < maxNumLights; ++lightIndex)
	{
		dynamicShadows[lightIndex].Push(new GeomShadowVolume(carGeom[0], &carBodyShadowMesh));
		for(ulong i = 1; i < 5; ++i)
			dynamicShadows[lightIndex].Push(new GeomShadowVolume(carGeom[i], &carWheelShadowMesh));
	}
}

void CreateTableRamps()
{
	static const dReal offsetX = _R(-20);

	// general stuff
	static const oReal tableLength = _R(5);
	static const oReal tableHeight = _R(10);
	static const oReal tableRampWidth = _R(2.5);
	static const oReal tableRampLength = oRealSqrt(tableHeight * tableHeight / _R(16) + tableLength * tableLength) - _R(0.17);

	// ramp types
	static const oVector3 tableTop(tableLength + tableRampWidth, rampThickness, tableLength + 2 * tableRampWidth);
	static const oVector3 tableSupport(supportThickness, tableHeight, supportThickness);
	static const oVector3 tableRampCorner(tableRampWidth, rampThickness, tableRampWidth);
	static const oVector3 tableRamp(tableRampLength, rampThickness, tableRampWidth);

	// the table top and supports
	tableGeom[0] = dCreateBox(0, tableTop.x, tableTop.y, tableTop.z);
	dGeomSetPosition(tableGeom[0], offsetX - tableRampWidth / 2, tableHeight, 0);
	for(ulong i = 1; i < 5; ++i)
		tableGeom[i] = dCreateBox(0, tableSupport.x, tableSupport.y, tableSupport.z);
	oReal tmp = tableLength / _R(2);
    dGeomSetPosition(tableGeom[1], offsetX + tmp, tableSupport.y / 2,  tmp);
    dGeomSetPosition(tableGeom[2], offsetX - tmp, tableSupport.y / 2,  tmp);
    dGeomSetPosition(tableGeom[3], offsetX - tmp, tableSupport.y / 2, -tmp);
    dGeomSetPosition(tableGeom[4], offsetX + tmp, tableSupport.y / 2, -tmp);

	// the outer ramp corners
	tmp = (tableLength + tableRampWidth) / _R(2);
	tableGeom[5] = dCreateBox(0, tableRampCorner.x, tableRampCorner.y, tableRampCorner.z);
	dGeomSetPosition(tableGeom[5], offsetX-tmp, tableHeight * _R(0.25), tmp);
	tableGeom[6] = dCreateBox(0, tableRampCorner.x, tableRampCorner.y, tableRampCorner.z);
	dGeomSetPosition(tableGeom[6], offsetX-tmp, tableHeight * _R(0.5), -tmp);
	tableGeom[7] = dCreateBox(0, tableRampCorner.x, tableRampCorner.y, tableRampCorner.z);
	dGeomSetPosition(tableGeom[7], offsetX+tmp, tableHeight * _R(0.75), -tmp);
	tableGeom[8] = dCreateBox(0, tableRampCorner.x, tableRampCorner.y, tableRampCorner.z);
	dGeomSetPosition(tableGeom[8], offsetX+tmp, tableHeight, tmp);
	tableGeom[9] = dCreateBox(0, tableRampCorner.x, tableRampCorner.y, tableRampCorner.z);
	dGeomSetPosition(tableGeom[9], offsetX+tmp, tableHeight, -tmp);

	// the side ramps
	tmp = (tableLength + tableRampWidth) / _R(2);
	dMatrix3 r, s, t;
	dRFromAxisAndAngle(r, 0, 0, 1, atan(-tableHeight / (4 * tableLength)));
	tableGeom[10] = dCreateBox(0, tableRamp.x, tableRamp.y, tableRamp.z);
	dGeomSetRotation(tableGeom[10], r);
	dGeomSetPosition(tableGeom[10], offsetX, tableHeight / 8, tmp);
	dRFromAxisAndAngle(s, 0, 1, 0, -M_PI / 2);
	dMULTIPLY0_333(t, s, r);
	tableGeom[11] = dCreateBox(0, tableRamp.x, tableRamp.y, tableRamp.z);
	dGeomSetRotation(tableGeom[11], t);
	dGeomSetPosition(tableGeom[11], offsetX-tmp, tableHeight * 3 / 8, 0);
	dRFromAxisAndAngle(s, 0, 1, 0, -M_PI);
	dMULTIPLY0_333(t, s, r);
	tableGeom[12] = dCreateBox(0, tableRamp.x, tableRamp.y, tableRamp.z);
	dGeomSetRotation(tableGeom[12], t);
	dGeomSetPosition(tableGeom[12], offsetX, tableHeight * 5 / 8, -tmp);
	dRFromAxisAndAngle(s, 0, 1, 0, M_PI / 2);
	dMULTIPLY0_333(t, s, r);
	tableGeom[13] = dCreateBox(0, tableRamp.x, tableRamp.y, tableRamp.z);
	dGeomSetRotation(tableGeom[13], t);
	dGeomSetPosition(tableGeom[13], offsetX+tmp, tableHeight * 7 / 8, 0);

	// the ramp supports
	for(ulong i = 14; i < 18; ++i)
		tableGeom[i] = dCreateBox(0, tableSupport.x, tableSupport.y, tableSupport.z);
	tmp = tableLength / _R(2) + tableRampWidth;
    dGeomSetPosition(tableGeom[14], offsetX+tmp, tableSupport.y / 2,  tmp);
    dGeomSetPosition(tableGeom[15], offsetX-tmp, tableSupport.y / 2,  tmp);
    dGeomSetPosition(tableGeom[16], offsetX-tmp, tableSupport.y / 2, -tmp);
    dGeomSetPosition(tableGeom[17], offsetX+tmp, tableSupport.y / 2, -tmp);
	
	// add to the environment geometry group
	for(ulong i = 0; i < 18; ++i)
		dGeomGroupAdd(envGeomGroup, tableGeom[i]);

	// ramp meshes 
	tableMeshes[0].CreateBox(tableTop, oVector3(tableRampWidth, tableTop.y, tableRampWidth));
	tableMeshes[1].CreateBox(tableSupport);
	tableMeshes[2].CreateBox(tableRampCorner);
	tableMeshes[3].CreateBox(tableRamp);
	staticMeshes.Push(new GeomMesh(tableGeom[0], &tableMeshes[0]));
	for(ulong i = 1; i < 5; ++i)
		staticMeshes.Push(new GeomMesh(tableGeom[i], &tableMeshes[1]));
	for(ulong i = 5; i < 10; ++i)
		staticMeshes.Push(new GeomMesh(tableGeom[i], &tableMeshes[2]));
	for(ulong i = 10; i < 14; ++i)
		staticMeshes.Push(new GeomMesh(tableGeom[i], &tableMeshes[3]));
	for(ulong i = 14; i < 18; ++i)
		staticMeshes.Push(new GeomMesh(tableGeom[i], &tableMeshes[1]));
	
	// ramp shadows
	tableShadowMeshes[0].CreateBox(tableTop);
	tableShadowMeshes[1].CreateBox(tableSupport);
	tableShadowMeshes[2].CreateBox(tableRampCorner);
	tableShadowMeshes[3].CreateBox(tableRamp);
	for(ulong lightIndex = 0; lightIndex < maxNumLights; ++lightIndex)
	{
		staticShadows[lightIndex].Push(new GeomShadowVolume(tableGeom[0], &tableShadowMeshes[0]));
        for(ulong i = 1; i < 5; ++i)
			staticShadows[lightIndex].Push(new GeomShadowVolume(tableGeom[i], &tableShadowMeshes[1]));
        for(ulong i = 5; i < 10; ++i)
			staticShadows[lightIndex].Push(new GeomShadowVolume(tableGeom[i], &tableShadowMeshes[2]));
        for(ulong i = 10; i < 14; ++i)
			staticShadows[lightIndex].Push(new GeomShadowVolume(tableGeom[i], &tableShadowMeshes[3]));
        for(ulong i = 14; i < 18; ++i)
			staticShadows[lightIndex].Push(new GeomShadowVolume(tableGeom[i], &tableShadowMeshes[1]));
	}
}

void CreateJumpRamp()
{
	static const dReal offsetX = _R(20);

	static const oReal rampLength = _R(10);
	static const oReal rampWidth = _R(5);
	static const oReal rampHeight = _R(4);
	static const oReal rampSpacing = _R(9);
	
	static const oReal rampFootprint = oRealSqrt(rampLength * rampLength - rampHeight * rampHeight);
	
	static const oVector3 rampScale(rampLength, rampThickness, rampWidth);	
	static const oVector3 supportScale(supportThickness, rampHeight - rampThickness, supportThickness);

	// create the ramps
	dReal tmp = (rampFootprint + rampSpacing) / _R(2);
	dMatrix3 r, s, t;
	jumpGeom[0] = dCreateBox(0, rampScale.x, rampScale.y, rampScale.z);
	dRFromAxisAndAngle(r, 0, 0, 1, asin(rampHeight / rampLength));
	dGeomSetRotation(jumpGeom[0], r);
	dGeomSetPosition(jumpGeom[0], offsetX - tmp, (rampHeight - rampThickness) / _R(2), _R(0));
	jumpGeom[1] = dCreateBox(0, rampScale.x, rampScale.y, rampScale.z);
	dRFromAxisAndAngle(s, 0, 1, 0, M_PI / _R(2));
	dMULTIPLY0_333(t, s, r);
	dGeomSetRotation(jumpGeom[1], t);
	dGeomSetPosition(jumpGeom[1], offsetX, (rampHeight - rampThickness) / _R(2), tmp);
	jumpGeom[2] = dCreateBox(0, rampScale.x, rampScale.y, rampScale.z);
	dRFromAxisAndAngle(s, 0, 1, 0, M_PI);
	dMULTIPLY0_333(t, s, r);
	dGeomSetRotation(jumpGeom[2], t);
	dGeomSetPosition(jumpGeom[2], offsetX + tmp, (rampHeight - rampThickness) / _R(2), _R(0));
	jumpGeom[3] = dCreateBox(0, rampScale.x, rampScale.y, rampScale.z);
	dRFromAxisAndAngle(s, 0, 1, 0, -M_PI / _R(2));
	dMULTIPLY0_333(t, s, r);
	dGeomSetRotation(jumpGeom[3], t);
	dGeomSetPosition(jumpGeom[3], offsetX, (rampHeight - rampThickness) / _R(2), -tmp);

	// support the ramps
	tmp = rampSpacing / _R(2) + supportThickness;
	dReal tmp2 = rampWidth / _R(2) - supportThickness;
	for(ulong i = 4; i < 12; ++i)
		jumpGeom[i] = dCreateBox(0, supportScale.x, supportScale.y, supportScale.z);
	dGeomSetPosition(jumpGeom[4], offsetX - tmp, supportScale.y / 2, -tmp2);
	dGeomSetPosition(jumpGeom[5], offsetX - tmp, supportScale.y / 2, tmp2);
	dGeomSetPosition(jumpGeom[6], offsetX + tmp, supportScale.y / 2, -tmp2);
	dGeomSetPosition(jumpGeom[7], offsetX + tmp, supportScale.y / 2, tmp2);
	dGeomSetPosition(jumpGeom[8], offsetX - tmp2, supportScale.y / 2, -tmp);
	dGeomSetPosition(jumpGeom[9], offsetX + tmp2, supportScale.y / 2, -tmp);
	dGeomSetPosition(jumpGeom[10], offsetX - tmp2, supportScale.y / 2, tmp);
	dGeomSetPosition(jumpGeom[11], offsetX + tmp2, supportScale.y / 2, tmp);

	// add to the environment geometry group
	for(ulong i = 0; i < 12; ++i)
		dGeomGroupAdd(envGeomGroup,jumpGeom[i]);

	// ramp meshes 
    jumpMeshes[0].CreateBox(rampScale);
	jumpMeshes[1].CreateBox(supportScale);
	for(ulong i = 0; i < 4; ++i)
		staticMeshes.Push(new GeomMesh(jumpGeom[i], &jumpMeshes[0]));
	for(ulong i = 4; i < 12; ++i)
		staticMeshes.Push(new GeomMesh(jumpGeom[i], &jumpMeshes[1]));

	// ramp shadows
	jumpShadowMeshes[0].CreateBox(rampScale);
	jumpShadowMeshes[1].CreateBox(supportScale);
	for(ulong lightIndex = 0; lightIndex < maxNumLights; ++lightIndex)
	{
		for(ulong i = 0; i < 4; ++i)
			staticShadows[lightIndex].Push(new GeomShadowVolume(jumpGeom[i], &jumpShadowMeshes[0]));
		for(ulong i = 4; i < 12; ++i)
			staticShadows[lightIndex].Push(new GeomShadowVolume(jumpGeom[i], &jumpShadowMeshes[1]));
	}	
}

void CreateBeachBalls()
{
	static const dReal offsetX = _R(-5);

	// general stuff
	static const oReal radius = _R(1);
	static const oReal spacing = _R(1.001);

	// the ball mass
	dMass m;
	dMassSetSphere(&m, 1, radius);
	dMassAdjust(&m, 2);

	// balls
	for(long k = 0, i = 0; k < 3; ++k)
	{
		for(long v = 0; v < 3; ++v)
		{
			for(long u = 0; u < 3; ++u, ++i)
			{
				ballGeom[i] = dCreateSphere(space, radius);
				ballBodies[i] = dBodyCreate(world);
				dBodySetMass(ballBodies[i], &m);
				dGeomSetBody(ballGeom[i], ballBodies[i]);
				dBodySetPosition(ballBodies[i], offsetX + _R(2 * u - 2) * spacing, _R(2 * k + 1) * spacing, _R(2 * v - 2) * spacing);
			}
		}
	}

	// guards
	for(ulong i = 0; i < 4; ++i)
		ballGuardGeom[i] = dCreateBox(0, 0.4, 0.4, 5);
	
	dMatrix3 r, s, t;
	dRFromAxisAndAngle(r, 0, 0, 1, M_PI / 4);
	dRFromAxisAndAngle(s, 0, 1, 0, M_PI / 2);
	dMULTIPLY0_333(t, s, r);

	dGeomSetRotation(ballGuardGeom[0], r);
	dGeomSetPosition(ballGuardGeom[0], offsetX + 2.8, 0, 0);
	dGeomSetRotation(ballGuardGeom[1], r);
	dGeomSetPosition(ballGuardGeom[1], offsetX - 2.8, 0, 0);
	dGeomSetRotation(ballGuardGeom[2], t);
	dGeomSetPosition(ballGuardGeom[2], offsetX, 0, 2.8);
	dGeomSetRotation(ballGuardGeom[3], t);
	dGeomSetPosition(ballGuardGeom[3], offsetX, 0, -2.8);

	for(ulong i = 0; i < 4; ++i)
		dGeomGroupAdd(envGeomGroup, ballGuardGeom[i]);

	// ball meshes
	ballMesh.CreateSphere(radius, 30, 15);
	ballShadowMesh.CreateSphere(radius, 30, 15);
	ballGuardMesh.CreateBox(oVector3(_R(0.4), _R(0.4), _R(5)));
	ballGuardShadowMesh.CreateBox(oVector3(_R(0.4), _R(0.4), _R(5)));

	for(ulong i = 0; i < 27; ++i)
		dynamicMeshes.Push(new GeomMesh(ballGeom[i], &ballMesh));
	for(ulong i = 0; i < 4; ++i)
		staticMeshes.Push(new GeomMesh(ballGuardGeom[i], &ballGuardMesh));

	for(ulong lightIndex = 0; lightIndex < maxNumLights; ++lightIndex)
	{
		for(ulong i = 0; i < 27; ++i)
			dynamicShadows[lightIndex].Push(new GeomShadowVolume(ballGeom[i], &ballShadowMesh));
		for(ulong i = 0; i < 4; ++i)
			staticShadows[lightIndex].Push(new GeomShadowVolume(ballGuardGeom[i], &ballGuardShadowMesh));
	}
}

// Modified by NateW for joystick support
void SetBuggyUI(double velocity, double steering)
{
	static const dReal steeringRate = M_PI * 4 / 3;
	static const dReal steeringLimit = M_PI / 6;

	dReal wheelVelocity = 12 * M_PI * velocity;// * 1.20; (sorry Nate, but this is waaay to fast for keyboarders) :)

	// a little bit of deadband seems to make it a bit easier to control
	if (fabs (steering) < 0.1)
		steering = 0;

	for(ulong i = 0; i < 4; ++i)
	{
		double desiredPosition = steering * ((i > 1) ? -1 : 1);
		double actualPosition = dJointGetHinge2Angle1(carJoints[i]);
		double steeringVelocity = (desiredPosition - actualPosition) * 10;

		dJointSetHinge2Param(carJoints[i], dParamHiStop, steeringLimit);
		dJointSetHinge2Param(carJoints[i], dParamLoStop, -steeringLimit);
		dJointSetHinge2Param(carJoints[i], dParamVel, steeringVelocity);
	}

	for(ulong i = 0; i < 4; ++i)
		dJointSetHinge2Param(carJoints[i], dParamVel2, ((i % 2) == 0) ? -wheelVelocity : wheelVelocity);
}

static void NearCallback(void* /*data*/, dGeomID o1, dGeomID o2)
{
	const ulong N = 6;
	dContact contacts[N];
	// get the contacts up to a maximum of N contacts
	const ulong n = dCollide(o1, o2, N, &contacts[0].geom, sizeof(dContact));
	for(ulong i = 0; i < n; ++i)
	{
        contacts[i].surface.mode = dContactBounce;
		contacts[i].surface.mu = 250;
		contacts[i].surface.bounce = 0.3;
		contacts[i].surface.bounce_vel = 0.05;
		dJointID c = dJointCreateContact(world, contactGroup, &contacts[i]);
		dJointAttach(c, dGeomGetBody(contacts[i].geom.g1), dGeomGetBody(contacts[i].geom.g2));
	}
}

class ComputationThread : public oRunnable
{
protected:
	oEvent stepSignal, workComplete;
	float frameTime;
	bool bQuit;

public:
	ComputationThread() : workComplete(true),  frameTime(0), bQuit(false)
	{
		workComplete.Signal();
	}

	void Step(const oReal& frameTime_) // caution NOT threadsafe
	{
        frameTime = frameTime_;
		workComplete.Reset();
		stepSignal.Signal();
	}

	HANDLE GetWorkCompleteEventHandle() { return workComplete; }

	virtual uint Run() throw()
	{
		for(;;)
		{
			workComplete.Signal();
			stepSignal.Wait();
			if(bQuit)
				return 0;

			// compute contact joints and forces
			dSpaceCollide(space, 0, &NearCallback);
			for(ulong i = 0; i < 16; ++i)
			{
				static const dReal u = -0.1;
				const dReal* pos = dBodyGetLinearVel(ballBodies[i]);
				dBodyAddForce(ballBodies[i], u * pos[0], u * pos[1], u * pos[2]);
				const dReal* rot = dBodyGetAngularVel(ballBodies[i]);
				dBodyAddTorque(ballBodies[i], u * rot[0], u * rot[1], u * rot[2]);
			}

			// step the world
			dWorldStep(world, frameTime);
			dJointGroupEmpty(contactGroup);
		}
	}

	virtual void Signal()
	{
		bQuit = true;
		stepSignal.Signal();
	}
};

int APIENTRY _tWinMain(HINSTANCE instance,
					   HINSTANCE /*prevInstance*/,
					   LPTSTR    /*cmdLine*/,
					   int       /*cmdShow*/)
{
	oGraphics graphics; // scoped here so that MessageBox works for caught exceptions (keeps the window alive)
	try 
	{
		// check we can get a decent idea of frame time
		graphics.GetPerformanceQueryResolution();

		// show the loader
		oGraphicsLoader loader;
		loader.SetMinPixelShaderVersion(0xffff); // effectively disallows pixel shaders
		if(!loader.Show(instance))
			return 0;

		// create the graphics device from the loader settings
		graphics.SetDepthStencilFormat(D3DFMT_D24S8);
		graphics.CreateFromLoader(loader, instance, _T("Shadowing Buggy Demo"), true);

		// create the DirectInput wrapper
		oInput input;
		input.Create(instance, graphics.GetTargetWindow());

		// load the GUI font
		oFont font;
		font.LoadFromFiles(graphics, _T("media/guifont.tga"), _T("media/guifont.w"));
		
		// precompute the gui transform
		oMatrix4x4 guiTransform;
		oCreateGUITransform(graphics.GetTargetWidth(), graphics.GetTargetHeight(), guiTransform);
		static const float textColour[] = { 1.0f, 1.0f, 1.0f, 1.0f };

		// create a GUI shader
		oVertexShader guiShader;
		oCreateGUIVertexShader(graphics, guiShader);

		// create a text buffer for the GUI text
		oTextBuffer staticText, dynamicText;
		staticText.Create(graphics, 500);
		dynamicText.Create(graphics, 300);
		staticText.Print(10, 10, _T("drive: up/down/left/right"), 0, font);
		staticText.Print(10, 25, _T("turbo: TAB"), 0, font);
		staticText.Print(10, 40, _T("flip upright: hold backspace"), 0, font);
		staticText.Print(10, 70, _T("toggle camera distance: C"), 0, font);
		staticText.Print(10, 85, _T("centre camera: space"), 0, font);
		staticText.Print(10, 100, _T("manual camera: hold control"), 0, font);
		staticText.Print(10, 115, _T("manual camera controls: W/S/A/D + up/down/left/right"), 0, font);
		staticText.Print(10, 145, _T("toggle number of lights: L"), 0, font);
		staticText.Print(10, 160, _T("rotate light 1 or 2: hold 1 or 2 (shift to rotate in reverse)"), 0, font);
		staticText.Print(10, 175, _T("show shadow volumes: hold F1 or F2"), 0, font);
		staticText.Print(10, 205, _T("toggle pause: P"), 0, font);
		staticText.Print(10, 220, _T("toggle help: H"), 0, font);

		// create the vertex shaders
		oVertexShader meshVS, ambientVS, shadowVS, shadowMeshVS;
		graphics.CreateVertexShader(meshVS, MeshVertex::decl, dwMeshVertexShader);
		graphics.CreateVertexShader(ambientVS, MeshVertex::decl, dwAmbientVertexShader);
		graphics.CreateVertexShader(shadowVS, oVector3Decl, dwShadowVertexShader);
		graphics.CreateVertexShader(shadowMeshVS, oVector3Decl, dwShadowmeshVertexShader);

		// load the textures
		HRESULT hr;
		oComPtr<IDirect3DTexture8>::Type grassTexture;
		hr = D3DXCreateTextureFromFileEx(
			graphics, 
			_T("media/grass.jpg"),
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, // all mip levels
			0, 
			D3DFMT_R8G8B8, 
			D3DPOOL_MANAGED, 
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, 
			0, 
			0, 
			&grassTexture
		);
		if(FAILED(hr))
			throw _T("Couldn't load texture from \"media/grass.jpg\"");
		oComPtr<IDirect3DTexture8>::Type rockTexture;
		hr = D3DXCreateTextureFromFileEx(
			graphics, 
			_T("media/grey_2.jpg"),
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, // all mip levels
			0, 
			D3DFMT_R8G8B8, 
			D3DPOOL_MANAGED, 
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, 
			0, 
			0, 
			&rockTexture
		);
		if(FAILED(hr))
			throw _T("Couldn't load texture from \"media/grey_2.jpg\"");
		oComPtr<IDirect3DTexture8>::Type bananaTexture;
		hr = D3DXCreateTextureFromFileEx(
			graphics, 
			_T("media/banana.jpg"),
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, // all mip levels
			0, 
			D3DFMT_R8G8B8, 
			D3DPOOL_MANAGED, 
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, 
			0, 
			0, 
			&bananaTexture
		);
		if(FAILED(hr))
			throw _T("Couldn't load texture from \"media/banana.jpg\"");
		oComPtr<IDirect3DTexture8>::Type wheelTexture;
		hr = D3DXCreateTextureFromFileEx(
			graphics, 
			_T("media/wheel.bmp"),
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, // all mip levels
			0, 
			D3DFMT_R8G8B8, 
			D3DPOOL_MANAGED, 
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, 
			0, 
			0, 
			&wheelTexture
		);
		if(FAILED(hr))
			throw _T("Couldn't load texture from \"media/wheel.bmp\"");
		oComPtr<IDirect3DTexture8>::Type ballTexture;
		hr = D3DXCreateTextureFromFileEx(
			graphics, 
			_T("media/beachball.jpg"),
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, // all mip levels
			0, 
			D3DFMT_R8G8B8, 
			D3DPOOL_MANAGED, 
			D3DX_DEFAULT, 
			D3DX_DEFAULT, 
			0, 
			0, 
			0, 
			&ballTexture
		);
		if(FAILED(hr))
			throw _T("Couldn't load texture from \"media/beachball.jpg\"");

		// create the world entities
		CreateWorld();
		CreateBuggy();
		CreateTableRamps();
		CreateJumpRamp();
		CreateBeachBalls();

		// create the static and dynamic vertex and index buffers
		oVertexBuffer<MeshVertex> dynamicVertices;
		oVertexBuffer<MeshVertex> staticVertices;
		oIndexBuffer<ushort> staticIndices;
		graphics.CreateVertexBuffer(staticVertices, 2000);
		graphics.CreateVertexBuffer(dynamicVertices, 15000);
		graphics.CreateIndexBuffer(staticIndices, 35000);
		
		// preload all the vertex and index data
		{
			ushort vertexOffset = 0, indexOffset = 0;
			for(ulong i = 0; i < staticMeshes.GetSize(); ++i)
			{
				vertexOffset = staticMeshes[i]->GetVertices(staticVertices, vertexOffset);
				indexOffset = staticMeshes[i]->GetIndices(staticIndices, indexOffset);
			}
			vertexOffset = 0;
			for(ulong i = 0; i < dynamicMeshes.GetSize(); ++i)
			{
				vertexOffset = dynamicMeshes[i]->GetVertices(dynamicVertices, vertexOffset);
				indexOffset = dynamicMeshes[i]->GetIndices(staticIndices, indexOffset);
			}
		}

		// create vertex and index buffers for shadow volume meshes
        oVertexBuffer<oVector3> dynamicShadowVertices;
		oIndexBuffer<ushort> dynamicShadowIndices;
		oVertexBuffer<oVector3> staticShadowVertices;
		oIndexBuffer<ushort> staticShadowIndices;
		graphics.CreateVertexBuffer(dynamicShadowVertices, 20000);
		graphics.CreateIndexBuffer(dynamicShadowIndices, 40000);
		graphics.CreateVertexBuffer(staticShadowVertices, 2000);
		graphics.CreateIndexBuffer(staticShadowIndices, 3000);

		// preload the static shadow volumes
		for(ushort lightIndex = 0, vertexOffset = 0, indexOffset = 0; lightIndex < maxNumLights; ++lightIndex)
		{
			oVector3 basisA, basisB;
			lights[lightIndex] = baseLights[lightIndex];
			ComputeShadowVolumeBasis(lights[lightIndex], basisA, basisB);

			for(ulong i = 0; i < staticShadows[lightIndex].GetSize(); ++i)
			{
				staticShadows[lightIndex][i]->ComputeShadowVolume(lights[lightIndex], basisA, basisB);
				vertexOffset = staticShadows[lightIndex][i]->GetVertices(staticShadowVertices, vertexOffset);
				indexOffset = staticShadows[lightIndex][i]->GetIndices(staticShadowIndices, indexOffset);
			}
		}
		
		// set up a suitable projection for this window shape
		oProjection projection;
		projection.zNear = _R(0.1);
		projection.zFar = _R(150);
		projection.aspectRatio = graphics.GetTargetAspectRatio();
		projection.CreateTransform();

		// set the camera position
		BuggyFollower buggyFollower(_R(7), _R(1.5), _R(4)); // distance, height, omegaSq
		buggyFollower.followPosition = oVector3(_R(-30), _R(3), _R(10));

		oVector3 cameraPosition(_R(0), _R(8), _R(30));
		oQuaternion cameraOrientation(oVector3(_R(0), _R(1), _R(0)), oRealPi);

		// frame counters
		oReal totalFrameTime = _R(0), fps = _R(0);
		unsigned long totalFrames = 0;
		
		// create the speed brake
		HANDLE hTimer = CreateWaitableTimer(0, TRUE, 0);
		LARGE_INTEGER li;
		li.QuadPart = (LONGLONG)(-10000 * 20); // speed brake to 20ms per frame
		SetWaitableTimer(hTimer, &li, 0, 0, 0, TRUE);

		// some once-only device setup
		graphics->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
		graphics->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
		graphics->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_LINEAR);

		// create the computation thread
		ComputationThread worker;
		oThread workerThread;
		workerThread.Start(&worker);

		// keeps track of the input variable
		HANDLE frameHandles[] = { hTimer, worker.GetWorkCompleteEventHandle() };
		bool changingNumLights = false, showHelp = true, changingHelp = false;
		bool changingPaused = false, paused = false;
		bool changingCamera = false, nearCamera = true;

		for(;;)
		{
			if(graphics.PurgeMessageQueue())
			{
				worker.Signal();
				workerThread.Wait();
				return 0;
			}

			// wait for the timer to signal or a message arrival
			if(WaitForMultipleObjects(2, frameHandles, TRUE, INFINITE) == WAIT_OBJECT_0)
			{
				// reset the frame timer
				SetWaitableTimer(hTimer, &li, 0, 0, 0, TRUE);

				// get the frame time
				oReal frameTime = graphics.GetFrameTime();

				// do the frame counting stuff
				++totalFrames;
				totalFrameTime += frameTime;
				if(totalFrameTime > _R(0.25))
				{
					fps = _R(totalFrames) / totalFrameTime;
					// change the display text
					dynamicText.Clear();
					_TCHAR buffer[160];
					_stprintf(buffer, _T("%.2f fps"), _R(totalFrames) / totalFrameTime);
					dynamicText.Print(graphics.GetTargetWidth() - 120, 10, buffer, 0, font);
					_stprintf(buffer, _T("%d light(s)"), numLights);
					dynamicText.Print(graphics.GetTargetWidth() - 120, 25, buffer, 0, font);
					if(paused)
						dynamicText.Print(graphics.GetTargetWidth() - 120, 40, _T("Paused"), 0, font);
					// reset the counters
					totalFrameTime = _R(0);
					totalFrames = 0;
				}
				if(frameTime < _R(0.02)) // clamp the simulation frame time
					frameTime = _R(0.02);
				else if(frameTime > _R(0.03))
					frameTime = _R(0.03);

				// get the key states
				input.GetImmediateKeyboardData();

				if(input.IsKeyDown(DIK_P))
				{
					if(!changingPaused)
					{
						changingPaused = true;
						paused = !paused;
					}
				}
				else
					changingPaused = false;

				if(input.IsKeyDown(DIK_C))
				{
					if(!changingCamera)
					{
						changingCamera = true;
						nearCamera = !nearCamera;
						buggyFollower.followDistance = nearCamera ? _R(7) : _R(14);
						buggyFollower.followHeight = nearCamera ? _R(1.5) : _R(4);
					}
				}
				else
					changingCamera = false;

				if(paused || input.IsKeyDown(DIK_LCONTROL))
				{
					// move/rotate the camera
					const oReal moveSpeed = (input.IsKeyDown(DIK_LSHIFT) ? _R(5) : _R(15)) * frameTime;
					if(input.IsKeyDown(DIK_W))
						cameraPosition += oQuaternionRotation(cameraOrientation, oVector3(_R(0), _R(0), moveSpeed));
					if(input.IsKeyDown(DIK_S))
						cameraPosition += oQuaternionRotation(cameraOrientation, oVector3(_R(0), _R(0), -moveSpeed));
					if(input.IsKeyDown(DIK_D))
						cameraPosition += oQuaternionRotation(cameraOrientation, oVector3(moveSpeed, _R(0), _R(0)));
					if(input.IsKeyDown(DIK_A))
						cameraPosition += oQuaternionRotation(cameraOrientation, oVector3(-moveSpeed, _R(0), _R(0)));
					buggyFollower.followPosition = cameraPosition;
					buggyFollower.followVelocity = oVector3(_R(0), _R(0), _R(0));

					const oReal rotateSpeed = (oRealPi / _R(2)) * frameTime;
					if(input.IsKeyDown(DIK_RIGHT))
						cameraOrientation = oQuaternion(oVector3(_R(0), _R(1), _R(0)), rotateSpeed) * cameraOrientation;
					if(input.IsKeyDown(DIK_LEFT))
						cameraOrientation = oQuaternion(oVector3(_R(0), _R(1), _R(0)), -rotateSpeed) * cameraOrientation;
					if(input.IsKeyDown(DIK_UP))
						cameraOrientation = cameraOrientation * oQuaternion(oVector3(_R(1), _R(0), _R(0)), rotateSpeed);
					if(input.IsKeyDown(DIK_DOWN))
						cameraOrientation = cameraOrientation * oQuaternion(oVector3(_R(1), _R(0), _R(0)), -rotateSpeed);
					cameraOrientation /= cameraOrientation.Norm();
				}
				else
				{
					// update the buggy from the UI	

					/////////////////////////////////////////////////////////////////////////////////////
					// Begin NateW alterations
					//SetBuggyUI((input.IsKeyDown(DIK_UP) ? 1 : 0) - (input.IsKeyDown(DIK_DOWN) ? 1 : 0),
					//	(input.IsKeyDown(DIK_RIGHT) ? 1 : 0) - (input.IsKeyDown(DIK_LEFT) ? 1 : 0));

					// Now uses joystick instead
					RefreshControlInput ();

					if (input.IsKeyDown(DIK_UP))
						g_ControlInput.m_rJoystickY = -1;
					if (input.IsKeyDown(DIK_DOWN))
						g_ControlInput.m_rJoystickY = 1;

					if (input.IsKeyDown(DIK_LEFT))
						g_ControlInput.m_rJoystickX = -1;
					if (input.IsKeyDown(DIK_RIGHT))
						g_ControlInput.m_rJoystickX = 1;

					// boost with TAB (this should be fun!)
					if(input.IsKeyDown(DIK_TAB))
						g_ControlInput.m_rJoystickY *= 1.5;

					SetBuggyUI(-g_ControlInput.m_rJoystickY, g_ControlInput.m_rJoystickX);

					// End NateW alterations
					/////////////////////////////////////////////////////////////////////////////////////

					if(input.IsKeyDown(DIK_BACKSPACE))
					{
						dBodyAddForce(carBodies[0], 0, 800, 0);
						dBodyAddRelTorque(carBodies[0], 70, 0, 0);
					}

					// update the follow camera
					buggyFollower.buggyPosition.CreateFromdVector3(dBodyGetPosition(carBodies[0]));
					buggyFollower.buggyVelocity.CreateFromdVector3(dBodyGetLinearVel(carBodies[0]));
					
					oQuaternion buggyOrientation(dBodyGetQuaternion(carBodies[0]));
					oVector3 buggyDirection = oQuaternionRotation(buggyOrientation, oVector3(_R(1), _R(0), _R(0)));
					buggyFollower.buggyDirection = oVector2(buggyDirection.x, buggyDirection.z);
					if(buggyFollower.buggyDirection.NormSquared() < _R(0.001))
					{
						oVector3 buggyDirection = oQuaternionRotation(buggyOrientation, oVector3(_R(0), _R(-1), _R(0)));
						buggyFollower.buggyDirection = oVector2(buggyDirection.x, buggyDirection.z);
					}
					buggyFollower.buggyDirection /= buggyFollower.buggyDirection.Norm();

					buggyFollower.UpdatePosition(frameTime, input.IsKeyDown(DIK_SPACE));
					cameraPosition = buggyFollower.followPosition;
					cameraOrientation.CreateFromLookAtPoint(buggyFollower.buggyPosition - buggyFollower.followPosition);
				}

				// change the number of lights
				if(input.IsKeyDown(DIK_L))
				{
					if(!changingNumLights)
					{
						changingNumLights = true;
						if(numLights == 1)
						{
							lightColour[0] = lightColour[1] = lightColour[2] = _R(0.6);
							ambientColour[0] = ambientColour[1] = ambientColour[2] = _R(0.3);
							numLights = 2;
						}
						else
						{
							lightColour[0] = lightColour[1] = lightColour[2] = _R(0.8);
							ambientColour[0] = ambientColour[1] = ambientColour[2] = _R(0.4);
							numLights = 1;
						}
					}
				}
				else
					changingNumLights = false;

				// toggle help
				if(input.IsKeyDown(DIK_H))
				{
					if(!changingHelp)
					{
						changingHelp = true;
						showHelp = !showHelp;
					}
				}
				else
					changingHelp = false;

				// update the lights
				const oReal lightRotateSpeed = (oRealPi / _R(4)) * frameTime;
				if(input.IsKeyDown(DIK_1))
				{
					lightRotations[0] = oQuaternion(oVector3(_R(0), _R(1), _R(0)), 
						input.IsKeyDown(DIK_LSHIFT) ? -lightRotateSpeed : lightRotateSpeed) * lightRotations[0];
					lightRotations[0] /= lightRotations[0].Norm();
				}
				if(input.IsKeyDown(DIK_2) && numLights > 1)
				{
					lightRotations[1] = oQuaternion(oVector3(_R(0), _R(1), _R(0)), 
						input.IsKeyDown(DIK_LSHIFT) ? -lightRotateSpeed : lightRotateSpeed) * lightRotations[1];
					lightRotations[1] /= lightRotations[1].Norm();
				}
				for(ushort lightIndex = 0; lightIndex < numLights; ++lightIndex)
					lights[lightIndex] = oQuaternionRotation(lightRotations[lightIndex], baseLights[lightIndex]);

				if(!paused)
				{
					// update the dynamic object vertex data
					for(ulong i = 0; i < dynamicMeshes.GetSize(); ++i)
						dynamicMeshes[i]->UpdateVertices(dynamicVertices);
				}

				// compute the dynamic shadow volumes
				if(!paused || input.IsKeyDown(DIK_1) || (input.IsKeyDown(DIK_2) && numLights > 1))
				{
					for(ushort vertexOffset = 0, indexOffset = 0, lightIndex = 0; lightIndex < numLights; ++lightIndex)
					{
						oVector3 basisA, basisB;
						ComputeShadowVolumeBasis(lights[lightIndex], basisA, basisB);

						for(ulong i = 0; i < 5; ++i)
						{
							dynamicShadows[lightIndex][i]->ComputeShadowVolume(lights[lightIndex], basisA, basisB);
							vertexOffset = dynamicShadows[lightIndex][i]->GetVertices(dynamicShadowVertices, vertexOffset);
							indexOffset = dynamicShadows[lightIndex][i]->GetIndices(dynamicShadowIndices, indexOffset);
						}
						for(ulong i = 5; i < dynamicShadows[lightIndex].GetSize(); ++i)
						{
							dynamicShadows[lightIndex][i]->ComputeSphereShadowVolumeHack(
								lights[lightIndex], basisA, basisB, _R(1), 30
							);
							vertexOffset = dynamicShadows[lightIndex][i]->GetVertices(dynamicShadowVertices, vertexOffset);
							indexOffset = dynamicShadows[lightIndex][i]->GetIndices(dynamicShadowIndices, indexOffset);
						}
					}
				}

				// redo the static stuff too if we moved the light
				if(input.IsKeyDown(DIK_1) || (input.IsKeyDown(DIK_2) && numLights > 1))
				{
					// load the static shadow volumes
					for(ushort lightIndex = 0, vertexOffset = 0, indexOffset = 0; lightIndex < maxNumLights; ++lightIndex)
					{
						oVector3 basisA, basisB;
						ComputeShadowVolumeBasis(lights[lightIndex], basisA, basisB);

						for(ulong i = 0; i < staticShadows[lightIndex].GetSize(); ++i)
						{
							staticShadows[lightIndex][i]->ComputeShadowVolume(lights[lightIndex], basisA, basisB);
							vertexOffset = staticShadows[lightIndex][i]->GetVertices(staticShadowVertices, vertexOffset);
							indexOffset = staticShadows[lightIndex][i]->GetIndices(staticShadowIndices, indexOffset);
						}
					}
				}
				
				// start processing the next frame
				if(!paused)
					worker.Step(frameTime);

				// create a view-projection transform from the camera
				oMatrix4x4 cameraTransform;
				cameraTransform.CreateFromQuaternion(~cameraOrientation);
				cameraTransform.PreTranslate(-cameraPosition);
				cameraTransform = projection.GetTransform() * cameraTransform;

				// set up the device
				graphics->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
				graphics->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
				graphics->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
				graphics->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
				graphics->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE);
				
				graphics->SetVertexShaderConstant(TRANSFORM, &cameraTransform, 4);
				graphics->SetVertexShaderConstant(AMBIENTCOLOUR, ambientColour, 1);
			
				//////// begin rendering
				graphics->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, 0xff7fa0e0, 1.0f, 0);
				graphics->BeginScene();

				//////// ambient lighting pass
				graphics->SetVertexShader(ambientVS);
				graphics->SetIndices(staticIndices, 0);

				// draw the dynamic objects
				graphics.SetStreamSource(0, dynamicVertices);
				graphics->SetTexture(0, bananaTexture);
				dynamicMeshes[0]->DrawIndexedPrimitive(graphics);
				graphics->SetTexture(0, 0);
				graphics->SetTexture(0, wheelTexture);
				for(ulong i = 1; i < 5; ++i)
					dynamicMeshes[i]->DrawIndexedPrimitive(graphics);
				graphics->SetTexture(0, 0);
				graphics->SetTexture(0, ballTexture);
				for(ulong i = 5; i < dynamicMeshes.GetSize(); ++i)
					dynamicMeshes[i]->DrawIndexedPrimitive(graphics);
				graphics->SetTexture(0, 0);
				graphics.ReleaseStreamSource(0);
				
				// draw the static objects
				graphics.SetStreamSource(0, staticVertices);
				graphics->SetTexture(0, rockTexture);
				for(ulong i = 1; i < staticMeshes.GetSize(); ++i)
					staticMeshes[i]->DrawIndexedPrimitive(graphics);
				graphics->SetTexture(0, 0);
				graphics->SetTexture(0, grassTexture);
				staticMeshes[0]->DrawIndexedPrimitive(graphics);
				graphics->SetTexture(0, 0);
				graphics.ReleaseStreamSource(0);

				graphics->SetIndices(0, 0);

				//////// directional light pass(es) with stenciled shadows
				for(ushort lightIndex = 0; lightIndex < numLights; ++lightIndex)
				{
					if(lightIndex != 0)
						graphics->Clear(0, 0, D3DCLEAR_STENCIL, 0, 0, 0);

					graphics->SetVertexShader(shadowVS);

					// backfaces incrementing stencil
					graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
					graphics->SetRenderState(D3DRS_STENCILENABLE, TRUE);

					graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);		//
					graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ZERO);	// horrible ATI hack
					graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);	//
					graphics->SetRenderState(D3DRS_COLORWRITEENABLE, 0);

					graphics->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_INCR);
					graphics->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

					graphics.SetStreamSource(0, dynamicShadowVertices);
					graphics->SetIndices(dynamicShadowIndices, 0);
					for(ulong i = 0; i < dynamicShadows[lightIndex].GetSize(); ++i)
						dynamicShadows[lightIndex][i]->DrawIndexedPrimitive(graphics);
					graphics->SetIndices(0, 0);
					graphics.ReleaseStreamSource(0);
					graphics.SetStreamSource(0, staticShadowVertices);
					graphics->SetIndices(staticShadowIndices, 0);
					for(ulong i = 0; i < staticShadows[lightIndex].GetSize(); ++i)
						staticShadows[lightIndex][i]->DrawIndexedPrimitive(graphics);
					graphics->SetIndices(0, 0);
					graphics.ReleaseStreamSource(0);

					// frontfaces decrementing stencil
					graphics->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_DECR);
					graphics->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

					graphics.SetStreamSource(0, dynamicShadowVertices);
					graphics->SetIndices(dynamicShadowIndices, 0);
					for(ulong i = 0; i < dynamicShadows[lightIndex].GetSize(); ++i)
						dynamicShadows[lightIndex][i]->DrawIndexedPrimitive(graphics);
					graphics->SetIndices(0, 0);
					graphics.ReleaseStreamSource(0);
					graphics.SetStreamSource(0, staticShadowVertices);
					graphics->SetIndices(staticShadowIndices, 0);
					for(ulong i = 0; i < staticShadows[lightIndex].GetSize(); ++i)
						staticShadows[lightIndex][i]->DrawIndexedPrimitive(graphics);
					graphics->SetIndices(0, 0);
					graphics.ReleaseStreamSource(0);

					graphics->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
					graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
					graphics->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALPHA | D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE);

					//////// directional lighting passes
					graphics->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_EQUAL);
					graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_EQUAL);

					graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
					graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
					graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

					graphics->SetVertexShader(meshVS);
					graphics->SetIndices(staticIndices, 0);
					
					graphics->SetVertexShaderConstant(LIGHTCOLOUR, lightColour, 1);
					float lightDirectionTemp[] = { lights[lightIndex].x, lights[lightIndex].y, lights[lightIndex].z, _R(0) };
					graphics->SetVertexShaderConstant(LIGHTDIR, lightDirectionTemp, 1);

					// draw the dynamic objects
					graphics.SetStreamSource(0, dynamicVertices);
					graphics->SetTexture(0, bananaTexture);
					dynamicMeshes[0]->DrawIndexedPrimitive(graphics);
					graphics->SetTexture(0, 0);
					graphics->SetTexture(0, wheelTexture);
					for(ulong i = 1; i < 5; ++i)
						dynamicMeshes[i]->DrawIndexedPrimitive(graphics);
					graphics->SetTexture(0, 0);
					graphics->SetTexture(0, ballTexture);
					for(ulong i = 5; i < dynamicMeshes.GetSize(); ++i)
						dynamicMeshes[i]->DrawIndexedPrimitive(graphics);
					graphics->SetTexture(0, 0);
					graphics.ReleaseStreamSource(0);
					
					// draw the static objects
					graphics.SetStreamSource(0, staticVertices);
					graphics->SetTexture(0, rockTexture);
					for(ulong i = 1; i < staticMeshes.GetSize(); ++i)
						staticMeshes[i]->DrawIndexedPrimitive(graphics);
					graphics->SetTexture(0, 0);
					graphics->SetTexture(0, grassTexture);
					staticMeshes[0]->DrawIndexedPrimitive(graphics);
					graphics->SetTexture(0, 0);
					graphics.ReleaseStreamSource(0);

					graphics->SetIndices(0, 0);

					graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL); // default
					graphics->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_ALWAYS); // default

					graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
					graphics->SetRenderState(D3DRS_STENCILENABLE, FALSE);
					graphics->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
				}

				//////// render the volumes (if necessary)
				graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);

				graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
				graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
				graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_SRCCOLOR);

				graphics->SetVertexShader(ambientVS);
				for(ushort lightIndex = 0; lightIndex < numLights; ++lightIndex)
				{
                    if(input.IsKeyDown(volumeKeys[lightIndex]))
					{
						graphics->SetVertexShaderConstant(AMBIENTCOLOUR, volumeFilter + (lightIndex*4), 1);
						// render the volumes
						graphics.SetStreamSource(0, dynamicShadowVertices);
						graphics->SetIndices(dynamicShadowIndices, 0);
						for(ulong i = 0; i < dynamicShadows[lightIndex].GetSize(); ++i)
							dynamicShadows[lightIndex][i]->DrawIndexedPrimitive(graphics);
						graphics->SetIndices(0, 0);
						graphics.ReleaseStreamSource(0);
						graphics.SetStreamSource(0, staticShadowVertices);
						graphics->SetIndices(staticShadowIndices, 0);
						for(ulong i = 0; i < staticShadows[lightIndex].GetSize(); ++i)
							staticShadows[lightIndex][i]->DrawIndexedPrimitive(graphics);
						graphics->SetIndices(0, 0);
						graphics.ReleaseStreamSource(0);
					}
				}
				graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
				
				//////// render the text
				graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_ALWAYS);
				graphics->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);

				graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
				graphics->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
				graphics->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

				graphics->SetVertexShader(guiShader);

				graphics->SetVertexShaderConstant(GUI_TRANSFORM, &guiTransform, 4);
				graphics->SetVertexShaderConstant(GUI_COLOUR, textColour, 1);

				graphics->SetTexture(0, font.GetTexture());
				graphics->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
				graphics->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
				graphics->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
				graphics->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);

				if(showHelp) staticText.DrawPrimitive(graphics);
				dynamicText.DrawPrimitive(graphics);

				graphics->SetTexture(0, 0);

				graphics->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);

				graphics->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
				graphics->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
	
				//////// end rendering
				graphics->EndScene();
				graphics->Present(0, 0, 0, 0);			
			}
		}
	}
	catch(_TCHAR* excuse)
	{
		// show the excuse before ritual suicide
		MessageBox(0, excuse, _T("OpenEngine: Commencing harakiri"), MB_OK | MB_ICONERROR | MB_TASKMODAL);
	}

	return 0;
}
