#include "stdafx.h"
#include "shadowmesh.h"

void ComputeShadowVolumeBasis(const oVector3& light, oVector3& basisA, oVector3& basisB) throw()
{
	// completes an orthonormal basis using light as the first vector
	basisA = (light.z < light.x) ? oVector3(_R(0), _R(0), _R(1)) : oVector3(_R(1), _R(0), _R(0));
	basisA -= light * oDotProduct(light, basisA);
	basisA /= basisA.Norm();
	basisB = oCrossProduct(light, basisA);
	basisB /= basisB.Norm();
	// ensure the basis (basisA, light, basisB) is left-handed (so that the convex hulls are clockwise)
	if((oCrossProduct(basisA, light) - basisB).NormSquared() > _R(1)) // should be 0 or 2
		basisB = -basisB;
}

ulong oVector3Decl[] = {
	D3DVSD_STREAM(0), 
	D3DVSD_REG(0, D3DVSDT_FLOAT3), // position
	D3DVSD_END() 
};

inline oVector3 GroundProjection(const oVector3& v, const oVector3& light) throw()
{
	// projects a vector onto the z=-1 plane in the light direction
	oReal t = (v.y + _R(1)) / light.y;
	return oVector3(v.x - t * light.x, _R(-1), v.z - t * light.z);
}

ShadowMesh::ShadowMesh() : numVertices(0) {}

void ShadowMesh::CreateBox(const oVector3& length)
{
	// allocate
	numVertices = 8;
	oAlloc(vertices, numVertices);
	// fill
	for(ulong i = 0; i < numVertices; ++i)
	{
		vertices[i].x = (i & 0x01) ? _R(-0.5) : _R(0.5);
		vertices[i].y = (i & 0x02) ? _R(-0.5) : _R(0.5);
		vertices[i].z = (i & 0x04) ? _R(-0.5) : _R(0.5);
		vertices[i] *= length;
	}
}

void ShadowMesh::CreateCylinder(const oReal& radius, const oReal& length, const ushort& segments)
{
	// allocate
	numVertices = 2 * segments;
	oAlloc(vertices, numVertices);
	// compute the radii
	oArrayPtr<oReal>::Type s = new oReal[segments + 1];
	oArrayPtr<oReal>::Type c = new oReal[segments + 1];
	for(ushort i = 0; i < segments; ++i)
	{
		s[i] = oRealSin(_R(2) * oRealPi * _R(i) / _R(segments));
		c[i] = oRealCos(_R(2) * oRealPi * _R(i) / _R(segments));
	}
	// fill
	oVector3 scale(radius, length, radius);
	for(ushort i = 0, j = segments; i < segments; ++i, ++j)
	{
		vertices[i] = oVector3(s[i], _R(0.5), c[i]) * scale;
		vertices[j] = oVector3(s[i], _R(-0.5), c[i]) * scale;
	}
}

void ShadowMesh::CreateSphere(const oReal& radius, const ushort& segments, const ushort& levels)
{
	if(levels < 2)
		throw _T("ShadowMesh::CreateSphere(): Invalid number of levels requested");
    // allocate
	numVertices = (levels - 1) * segments + 2;
	oAlloc(vertices, numVertices);
	// fill vertices
	oVector3* buffer = vertices;
	*buffer++ = oVector3(_R(0), radius, _R(0));
    for(ushort v = 1; v < levels; ++v)
	{
		for(ushort u = 0; u < segments; ++u)
		{
			oReal s = oRealSin(_R(v) * oRealPi / _R(levels));
			oVector3 temp(
				s * oRealSin(_R(2) * oRealPi * _R(u) / _R(segments)), 
				oRealCos(_R(v) * oRealPi / _R(levels)), 
				s * oRealCos(_R(2) * oRealPi * _R(u) / _R(segments))
			);
            *buffer++ = radius * temp;
		}
	}
	*buffer++ = oVector3(_R(0), -radius, _R(0));
}

ShadowVolume::ShadowVolume() : numVertices(0) {}

ShadowVolume::ShadowVolume(const ShadowMesh* const& mesh_) : numVertices(0)
{
	Create(mesh_);
}

void ShadowVolume::Create(const ShadowMesh* const& mesh_)
{
	// allocate all the required memory
	mesh = mesh_;
	oAlloc(hull, mesh->GetVertexCount());
	oAlloc(hullIndices, mesh->GetVertexCount() + 1);
	oAlloc(vertices, 2 * mesh->GetVertexCount() + 2);
}

void ShadowVolume::ComputeShadowVolume(const oMatrix4x4& transform, const oVector3& light, const oVector3& basisA, const oVector3& basisB) throw()
{
	// project onto the plane normal to the light
	for(ushort i = 0; i < mesh->GetVertexCount(); ++i)
	{
		oVector3 temp = oTransformAs3x3(transform, mesh->GetVertex(i));
		hull[i] = oVector2(oDotProduct(temp, basisA), oDotProduct(temp, basisB));
	}

	// compute the convex hull on this plane
	numVertices = oQuickHull(hull, mesh->GetVertexCount(), _R(0.005) * _R(0.005), hullIndices);

	// extrude the shadow volume from the planar convex hull
	vertices[0] = transform.GetTranslation();
    for(ushort i = 0, j = 1; i < numVertices; ++i, ++j)
        vertices[j] = transform * mesh->GetVertex(hullIndices[i]);
	for(ushort i = 0, j = numVertices + 1; i <= numVertices; ++i, ++j)
		vertices[j] = GroundProjection(vertices[i], light);

}

void ShadowVolume::ComputeSphereShadowVolumeHack(const oMatrix4x4& transform, const oVector3& light, const oVector3& basisA, const oVector3& basisB, const oReal& radius, const ushort& segments) throw()
{
	numVertices = segments;

	// construct a circle in the plane normal to the light
	vertices[0] = transform.GetTranslation();
	for(ushort i = 0, j = 1; i < numVertices; ++i, ++j)
		vertices[j] = transform.GetTranslation() 
			+ radius * oRealSin(_R(2) * oRealPi * _R(i) / _R(numVertices)) * basisA 
			+ radius * oRealCos(_R(2) * oRealPi * _R(i) / _R(numVertices)) * basisB
			- radius * _R(0.1) * light;
	for(ushort i = 0, j = numVertices + 1; i <= numVertices; ++i, ++j)
		vertices[j] = GroundProjection(vertices[i], light);
}

void ShadowVolume::GetVertices(oVector3* const& buffer) const throw()
{
	memcpy(buffer, vertices, GetVertexCount() * sizeof(oVector3));
}

void ShadowVolume::GetIndices(ushort* const& buffer, const ushort& offset) const throw()
{
	// construct the index buffer to render the shadow volume
	ulong idx = 0;
	buffer[idx++] = offset; // the lid fan
	for(ushort i = 1; i <= numVertices; ++i)
		buffer[idx++] = offset + i;
	buffer[idx++] = offset + 1;
	buffer[idx++] = offset + numVertices + 1; // the base fan
	for(ushort i = 2 * numVertices + 1; i > numVertices + 1; --i)
		buffer[idx++] = offset + i;
	buffer[idx++] = offset + 2 * numVertices + 1;
	for(ushort i = 1; i <= numVertices; ++i) // the sides strip
	{
		buffer[idx++] = offset + i;
		buffer[idx++] = offset + i + numVertices + 1;
	}
	buffer[idx++] = offset + 1;
	buffer[idx++] = offset + numVertices + 2;
}

void ShadowVolume::DrawIndexedPrimitive(IDirect3DDevice8* const& device, const ushort& vertexOffset, const ushort& indexOffset) const throw()
{
	device->DrawIndexedPrimitive(
		D3DPT_TRIANGLEFAN, 
		vertexOffset, numVertices + 1, 
		indexOffset, numVertices
	);
	device->DrawIndexedPrimitive(
		D3DPT_TRIANGLEFAN, 
		vertexOffset + numVertices + 1, numVertices + 1, 
		indexOffset + numVertices + 2, numVertices
	);
	device->DrawIndexedPrimitive(
		D3DPT_TRIANGLESTRIP, 
		vertexOffset, 2 * numVertices + 2, 
		indexOffset + 2 * numVertices + 4, 2 * numVertices
	);
}

dGeomID geom;
ushort vertexOffset, indexOffset;

GeomShadowVolume::GeomShadowVolume() : geom(0), vertexOffset(0), indexOffset(0) {}
    
GeomShadowVolume::GeomShadowVolume(const dGeomID& geom_, const ShadowMesh* const& mesh_) 
  : ShadowVolume(mesh_), vertexOffset(0), indexOffset(0)
{
	geom = geom_;
}
	
void GeomShadowVolume::Create(const dGeomID& geom_, const ShadowMesh* const& mesh_)
{
	geom = geom_;
	ShadowVolume::Create(mesh_);
}

void GeomShadowVolume::ComputeShadowVolume(const oVector3& light, const oVector3& basisA, const oVector3& basisB) throw()
{
	// compute the mesh and normal transform
	oMatrix4x4 transform(dGeomGetRotation(geom));
	transform.PostTranslate(oVector3(dGeomGetPosition(geom)));
	// compute the shadow volume
	ShadowVolume::ComputeShadowVolume(transform, light, basisA, basisB);
}

void GeomShadowVolume::ComputeSphereShadowVolumeHack(const oVector3& light, const oVector3& basisA, const oVector3& basisB, const oReal& radius, const ushort& segments) throw()
{
	// compute the mesh and normal transform
	oMatrix4x4 transform(dGeomGetRotation(geom));
	transform.PostTranslate(oVector3(dGeomGetPosition(geom)));
	// compute the shadow volume
	ShadowVolume::ComputeSphereShadowVolumeHack(transform, light, basisA, basisB, radius, segments);
}

void GeomShadowVolume::DrawIndexedPrimitive(IDirect3DDevice8* const& device)
{
	ShadowVolume::DrawIndexedPrimitive(device, vertexOffset, indexOffset);
}
