#include "openengine.h"

oVector3::oVector3() throw() : x(_R(0)), y(_R(0)), z(_R(0)) {}

oVector3::oVector3(const oReal& x_, const oReal& y_, const oReal& z_) throw() : x(x_), y(y_), z(z_) {}

oVector3::oVector3(const oVector3& v) throw() : x(v.x), y(v.y), z(v.z) {}

oVector3& oVector3::operator=(const oVector3& v) throw()
{
	x = v.x;
	y = v.y;
	z = v.z;
	return *this;
}

oReal oVector3::Norm() const throw()
{
	return oRealSqrt(NormSquared());
}

oReal oVector3::NormSquared() const throw()
{
	return x * x + y * y + z * z;
}

oVector3 oVector3::operator-() const throw()
{
	return oVector3(-x, -y, -z);
}

const oVector3& oVector3::operator+=(const oVector3& v) throw()
{
	x += v.x;
	y += v.y;
	z += v.z;
	return *this;
}

const oVector3& oVector3::operator-=(const oVector3& v) throw()
{
	x -= v.x;
	y -= v.y;
	z -= v.z;
	return *this;
}

const oVector3& oVector3::operator*=(const oVector3& v) throw()
{
	x *= v.x;
	y *= v.y; 
	z *= v.z;
	return *this;
}

const oVector3& oVector3::operator*=(const oReal& r) throw()
{
	x *= r;
	y *= r;
	z *= r;
	return *this;
}

const oVector3& oVector3::operator/=(const oReal& r) throw()
{
	x /= r;
	y /= r;
	z /= r;
	return *this;
}

oVector3 operator+(const oVector3& l, const oVector3& r) throw()
{
	return oVector3(l.x + r.x, l.y + r.y, l.z + r.z);
}

oVector3 operator-(const oVector3& l, const oVector3& r) throw()
{
	return oVector3(l.x - r.x, l.y - r.y, l.z - r.z);
}

oVector3 operator*(const oVector3& l, const oVector3& r) throw()
{
	return oVector3(l.x * r.x, l.y * r.y, l.z * r.z);
}

oVector3 operator*(const oReal& l, const oVector3& r) throw()
{
	return oVector3(l * r.x, l * r.y, l * r.z);
}

oVector3 operator*(const oVector3& l, const oReal& r) throw()
{
	return oVector3(l.x * r, l.y * r, l.z * r);
}

oVector3 operator/(const oVector3& l, const oReal& r) throw()
{
	return oVector3(l.x / r, l.y / r, l.z / r);
}

oReal oDotProduct(const oVector3& l, const oVector3& r) throw()
{
	return l.x * r.x + l.y * r.y + l.z * r.z;
}

oVector3 oCrossProduct(const oVector3& l, const oVector3& r) throw()
{
	return oVector3(l.y * r.z - l.z * r.y, 
				    l.z * r.x - l.x * r.z, 
				    l.x * r.y - l.y * r.x);
}

oLine3::oLine3() throw() : p(), d(_R(0), _R(0), _R(1)) {}

oLine3::oLine3(const oVector3& p_, const oVector3& d_) throw() : p(p_), d(d_) {}

oQuaternion::oQuaternion() throw() : w(_R(1)), x(_R(0)), y(_R(0)), z(_R(0)) {}

oQuaternion::oQuaternion(const oReal& w_, const oReal& x_, const oReal& y_, const oReal& z_) throw()
  : w(w_), x(x_), y(y_), z(z_) 
{
}

oQuaternion::oQuaternion(const oVector3& axis, const oReal& angle)
{
	CreateFromAxisAngle(axis, angle);
}

oQuaternion::oQuaternion(const oMatrix4x4& m) throw()
{
	CreateFromMatrix(m);
}

void oQuaternion::CreateFromMatrix(const oMatrix4x4& m) throw()
{
	// get w from the trace
	oReal tmp = m.e11 + m.e22 + m.e33 + _R(1);
	if(tmp < _R(0)) tmp = _R(0);
	w = oRealSqrt(tmp) / _R(2);
	// get x, y, z from the body
	tmp	= w * _R(4);
	x = (m.e32 - m.e23) / tmp;
	y = (m.e13 - m.e31) / tmp;
	z = (m.e21 - m.e12) / tmp;
}

oReal oQuaternion::Norm() const throw()
{
	return oRealSqrt(NormSquared());
}

oReal oQuaternion::NormSquared() const throw()
{
	return w * w + x * x + y * y + z * z;
}

oQuaternion oQuaternion::operator~() const throw()
{
	return oQuaternion(w, -x, -y, -z);
}

const oQuaternion& oQuaternion::operator/=(const oReal& r)
{
	w /= r;
	x /= r;
	y /= r;
	z /= r;
	return *this;
}

void oQuaternion::CreateFromAxisAngle(const oVector3& axis, const oReal& angle) throw()
{
	// cache angle / 2
	w = angle * _R(0.5);
	// cache sin(angle / 2)
	x = oRealSin(w);
	y = x * axis.y;
	z = x * axis.z;
	x *= axis.x;
	// store cos(angle / 2)
	w = oRealCos(w);
}

oQuaternion::oQuaternion(const oVector3& lookat) throw()
{
	CreateFromLookAtPoint(lookat);
}

void oQuaternion::CreateFromLookAtPoint(const oVector3& lookat) throw()
{
	static const oReal lengthEpsilon = _R(0.001); // 1mm reasonable?
	// get the y axis rotation
	oReal r = oRealSqrt(lookat.x * lookat.x + lookat.z * lookat.z);
	oReal ycos2 = _R(1), ysin2 = _R(0);
	if(r > lengthEpsilon)
	{
		oReal c = lookat.z / r;
		ycos2 = oRealSqrt((_R(1) + c) / 2);
		ysin2 = oRealSqrt((_R(1) - c) / 2);
		if(lookat.x < _R(0))
			ysin2 = -ysin2;
	}
	// get the (rotated frame) x axis rotation
	oReal d = oRealSqrt(r * r + lookat.y * lookat.y);
	oReal xcos2 = _R(1), xsin2 = _R(0);
	if(d > lengthEpsilon)
	{
		oReal c = r / d;
		xcos2 = oRealSqrt((_R(1) + c) / 2);
		xsin2 = oRealSqrt((_R(1) - c) / 2);
		if(lookat.y > _R(0))
			xsin2 = -xsin2;
	}
	// TODO: setup the rotation manually
	*this = oQuaternion(ycos2, _R(0), ysin2, _R(0)) * oQuaternion(xcos2, xsin2, _R(0), _R(0));
}

oQuaternion operator*(const oQuaternion& l, const oQuaternion& r) throw()
{
	return oQuaternion(l.w * r.w - (l.x * r.x + l.y * r.y + l.z * r.z), 
					   l.w * r.x + l.x * r.w + l.y * r.z - l.z * r.y,
					   l.w * r.y + l.y * r.w + l.z * r.x - l.x * r.z,
					   l.w * r.z + l.z * r.w + l.x * r.y - l.y * r.x);
}

oQuaternion operator/(const oQuaternion& l, const oReal& r) throw()
{
	return oQuaternion(l.w / r, l.x / r, l.y / r, l.z / r);
}

oQuaternion oQuaternionPower(const oQuaternion& q, const oReal& t) throw()
{
	oReal sinphi = oRealSqrt(q.x * q.x + q.y * q.y + q.z * q.z);
	if(sinphi < oRealEpsilon)
		return oQuaternion();
	// inverse the sin and return the power
	oReal phi = ((q.w < _R(0)) ? -t : t) * oRealInvSin(sinphi);
	oReal scale = oRealSin(phi) / sinphi;
	return oQuaternion(oRealCos(phi), q.x * scale, q.y * scale, q.z * scale);
}

oQuaternion oQuaternionSLERP(const oQuaternion& a, const oQuaternion& b, const oReal& t) throw()
{
	return oQuaternionPower(b * ~a, t) * a;
}

oVector3 oQuaternionRotation(const oQuaternion& q, const oVector3& v) throw()
{
	// form ([0,v].~q) directly
	oQuaternion t(
		v.x * q.x + v.y * q.y + v.z * q.z, 
		v.x * q.w + v.z * q.y - v.y * q.z,
		v.y * q.w + v.x * q.z - v.z * q.x, 			
		v.z * q.w + v.y * q.x - v.x * q.y
	); 
	// return the last three components of (q.[0,v].~q) as a vector
	return oVector3(
	   q.w * t.x + q.x * t.w + q.y * t.z - q.z * t.y,
	   q.w * t.y + q.y * t.w + q.z * t.x - q.x * t.z,
	   q.w * t.z + q.z * t.w + q.x * t.y - q.y * t.x
	);
}

oMatrix4x4::oMatrix4x4() throw()
  : e11(_R(1)), e12(_R(0)), e13(_R(0)), e14(_R(0)), 
	e21(_R(0)), e22(_R(1)), e23(_R(0)), e24(_R(0)), 
	e31(_R(0)), e32(_R(0)), e33(_R(1)), e34(_R(0)), 
	e41(_R(0)), e42(_R(0)), e43(_R(0)), e44(_R(1))
{
}

oMatrix4x4::oMatrix4x4(const oReal& e11_, const oReal& e12_, const oReal& e13_, const oReal& e14_, 
					   const oReal& e21_, const oReal& e22_, const oReal& e23_, const oReal& e24_, 
					   const oReal& e31_, const oReal& e32_, const oReal& e33_, const oReal& e34_, 
					   const oReal& e41_, const oReal& e42_, const oReal& e43_, const oReal& e44_) throw()
  : e11(e11_), e12(e12_), e13(e13_), e14(e14_), 
	e21(e21_), e22(e22_), e23(e23_), e24(e24_), 
	e31(e31_), e32(e32_), e33(e33_), e34(e34_), 
	e41(e41_), e42(e42_), e43(e43_), e44(e44_)
{
}

oMatrix4x4::oMatrix4x4(const oReal& e11_, const oReal& e12_, const oReal& e13_, 
					   const oReal& e21_, const oReal& e22_, const oReal& e23_, 
					   const oReal& e31_, const oReal& e32_, const oReal& e33_) throw()
  : e11(e11_), e12(e12_), e13(e13_), e14(_R(0)), 
	e21(e21_), e22(e22_), e23(e23_), e24(_R(0)), 
	e31(e31_), e32(e32_), e33(e33_), e34(_R(0)), 
	e41(_R(0)), e42(_R(0)), e43(_R(0)), e44(_R(1))
{
}

void oMatrix4x4::CreateIdentity() throw()
{
	e11 = e22 = e33 = e44 = _R(1);
	e12 = e13 = e14 = e41
		= e22 = e23 = e24 = e42
		= e32 = e33 = e34 = e43 = _R(0);
}

oMatrix4x4::oMatrix4x4(const oMatrix4x4& m) throw()
  : e11(m.e11), e12(m.e12), e13(m.e13), e14(m.e14), 
	e21(m.e21), e22(m.e22), e23(m.e23), e24(m.e24), 
	e31(m.e31), e32(m.e32), e33(m.e33), e34(m.e34), 
	e41(m.e41), e42(m.e42), e43(m.e43), e44(m.e44)
{
}

oMatrix4x4& oMatrix4x4::operator=(const oMatrix4x4& m) throw()
{
	e11 = m.e11; e12 = m.e12; e13 = m.e13; e14 = m.e14;
	e21 = m.e21; e22 = m.e22; e23 = m.e23; e24 = m.e24;
	e31 = m.e31; e32 = m.e32; e33 = m.e33; e34 = m.e34;
	e41 = m.e41; e42 = m.e42; e43 = m.e43; e44 = m.e44;
	return *this;
}

oMatrix4x4 oMatrix4x4::Transpose() const throw()
{
	return oMatrix4x4(e11, e21, e31, e41, 
					  e12, e22, e32, e42, 
					  e13, e23, e33, e43, 
					  e14, e24, e34, e44);
}

oMatrix4x4::oMatrix4x4(const oQuaternion& q) throw()
{
	CreateFromQuaternion(q);
}

void oMatrix4x4::CreateFromQuaternion(const oQuaternion& q) throw()
{
	oReal x2 = q.x + q.x;
	oReal y2 = q.y + q.y;
	oReal z2 = q.z + q.z;
	// fill in the non-diagonal entries first
	e12 = e21 = x2 * q.y;
	e23 = e32 = y2 * q.z;
	e31 = e13 = z2 * q.x;
	oReal t = q.w * x2;
	e23 -= t; e32 += t;
	t = q.w * y2;
	e31 -= t; e13 += t;
	t = q.w * z2;
	e12 -= t; e21 += t;
	// fill in the diagonals
	e11 = e22 = e33 = _R(1);
	t = x2 * q.x;
	e22 -= t; e33 -= t;
	t = y2 * q.y;
	e11 -= t; e33 -= t;
	t = z2 * q.z;
	e11 -= t; e22 -= t;
	// fill in the rest
	e14 = e24 = e34 = e41 = e42 = e43 = _R(0);
	e44 = _R(1);
}

void oMatrix4x4::CreateFromScale(const oVector3& v) throw()
{
	e11 = v.x; e12 = _R(0); e13 = _R(0); e14 = _R(0);
	e21 = _R(0); e22 = v.y; e23 = _R(0); e24 = _R(0);
	e31 = _R(0); e32 = _R(0); e33 = v.z; e34 = _R(0);
	e41 = _R(0); e42 = _R(0); e43 = _R(0); e44 = _R(1);
}

void oMatrix4x4::PreTranslate(const oVector3& v) throw()
{
	e14 += (e11 * v.x + e12 * v.y + e13 * v.z);
	e24 += (e21 * v.x + e22 * v.y + e23 * v.z);
	e34 += (e31 * v.x + e32 * v.y + e33 * v.z);
}

void oMatrix4x4::PostTranslate(const oVector3& v) throw()
{
	e14 += v.x;
	e24 += v.y;
	e34 += v.z;
}

void oMatrix4x4::PreScale(const oVector3& v) throw()
{
	e11 *= v.x; e12 *= v.y; e13 *= v.z;
	e21 *= v.x; e22 *= v.y; e23 *= v.z;
	e31 *= v.x; e32 *= v.y; e33 *= v.z;
	e41 *= v.x; e42 *= v.y; e43 *= v.z;
}

void oMatrix4x4::PostScale(const oVector3& v) throw()
{
	e11 *= v.x; e12 *= v.x; e13 *= v.x; e14 *= v.x;
	e21 *= v.y; e22 *= v.y; e23 *= v.y; e24 *= v.y;
	e31 *= v.z; e32 *= v.z; e33 *= v.z; e34 *= v.z;
}

oVector3 oMatrix4x4::GetTranslation() const throw()
{
	return oVector3(e14, e24, e34);
}

oMatrix4x4 operator*(const oMatrix4x4& l, const oMatrix4x4& r) throw()
{
	return oMatrix4x4(
		l.e11 * r.e11 + l.e12 * r.e21 + l.e13 * r.e31 + l.e14 * r.e41, 
		l.e11 * r.e12 + l.e12 * r.e22 + l.e13 * r.e32 + l.e14 * r.e42, 
		l.e11 * r.e13 + l.e12 * r.e23 + l.e13 * r.e33 + l.e14 * r.e43, 
		l.e11 * r.e14 + l.e12 * r.e24 + l.e13 * r.e34 + l.e14 * r.e44, 

		l.e21 * r.e11 + l.e22 * r.e21 + l.e23 * r.e31 + l.e24 * r.e41, 
		l.e21 * r.e12 + l.e22 * r.e22 + l.e23 * r.e32 + l.e24 * r.e42, 
		l.e21 * r.e13 + l.e22 * r.e23 + l.e23 * r.e33 + l.e24 * r.e43, 
		l.e21 * r.e14 + l.e22 * r.e24 + l.e23 * r.e34 + l.e24 * r.e44, 

		l.e31 * r.e11 + l.e32 * r.e21 + l.e33 * r.e31 + l.e34 * r.e41, 
		l.e31 * r.e12 + l.e32 * r.e22 + l.e33 * r.e32 + l.e34 * r.e42, 
		l.e31 * r.e13 + l.e32 * r.e23 + l.e33 * r.e33 + l.e34 * r.e43, 
		l.e31 * r.e14 + l.e32 * r.e24 + l.e33 * r.e34 + l.e34 * r.e44, 

		l.e41 * r.e11 + l.e42 * r.e21 + l.e43 * r.e31 + l.e44 * r.e41, 
		l.e41 * r.e12 + l.e42 * r.e22 + l.e43 * r.e32 + l.e44 * r.e42, 
		l.e41 * r.e13 + l.e42 * r.e23 + l.e43 * r.e33 + l.e44 * r.e43, 
		l.e41 * r.e14 + l.e42 * r.e24 + l.e43 * r.e34 + l.e44 * r.e44
	);
}

oVector3 operator*(const oMatrix4x4& l, const oVector3& r) throw()
{
	return oTransformAs3x3(l, r) + oVector3(l.e14, l.e24, l.e34);
}

oVector3 oTransformAs3x3(const oMatrix4x4& l, const oVector3& r) throw()
{
	return oVector3(l.e11 * r.x + l.e12 * r.y + l.e13 * r.z, 
					l.e21 * r.x + l.e22 * r.y + l.e23 * r.z, 
					l.e31 * r.x + l.e32 * r.y + l.e33 * r.z);
}

oProjection::oProjection() throw() : fovRatio(_R(2)), aspectRatio(_R(1)), zNear(_R(1)), zFar(_R(1000)) {}

void oProjection::CreateTransform() throw()
{
	// scale
	transform.e11 = fovRatio;
	transform.e22 = fovRatio * aspectRatio;
	// w-friendly perspective
	transform.e33 = zFar / (zFar - zNear);
	transform.e34 = -transform.e33 * zNear;
	transform.e43 = _R(1);
	// zero the rest
	transform.e12 = transform.e13 = transform.e14 = 
		transform.e21 = transform.e23 = transform.e24 = 
		transform.e31 = transform.e32 = 
		transform.e41 = transform.e42 = transform.e44 = _R(0);
}

oVector2::oVector2() throw() : x(_R(0)), y(_R(0)) {}

oVector2::oVector2(const oReal& x_, const oReal& y_) throw() : x(x_), y(y_) {}

oVector2::oVector2(const oVector2& v) throw() : x(v.x), y(v.y) {}

oVector2& oVector2::operator=(const oVector2& v) throw()
{
	x = v.x;
	y = v.y;
	return *this;
}

oReal oVector2::Norm() const throw()
{
	return oRealSqrt(NormSquared());
}

oReal oVector2::NormSquared() const throw()
{
	return x * x + y * y;
}

const oVector2& oVector2::operator*=(const oReal& r) throw()
{
	x *= r;
	y *= r;
	return *this;
}

const oVector2& oVector2::operator*=(const oVector2& v) throw()
{
	x *= v.x;
	y *= v.y;
	return *this;
}

const oVector2& oVector2::operator/=(const oReal& r) throw()
{
	x /= r; 
	y /= r;
	return *this;
}

oReal oDotProduct(const oVector2& l, const oVector2& r) throw()
{
	return l.x * r.x + l.y * r.y;
}

oPlane2::oPlane2() throw() : n(_R(1), _R(0)), d(0) {}

oPlane2::oPlane2(const oVector2& n_, const oReal& d) throw() : n(n_), d(d) {}

oPlane2::oPlane2(const oVector2& a, const oVector2& b) throw()
{
	CreateFromPoints(a, b);
}

oPlane2::oPlane2(const oPlane2& l) throw() : n(l.n), d(l.d) {}

oPlane2& oPlane2::operator=(const oPlane2& l) throw()
{
	n = l.n;
	d = l.d;
	return *this;
}

void oPlane2::CreateFromPoints(const oVector2& a, const oVector2& b) throw()
{
	// rotate the direction by pi/2 ccw (to describe the hull of a clockwise convex solid)
    n.x = a.y - b.y;
	n.y = b.x - a.x;
	d = oDotProduct(n, a);
}

oReal oPlane2::PointHeight(const oVector2& v) const throw()
{
	return oDotProduct(n, v) - d;
}

bool oPlane2::PointAboveTol2(const oVector2& v, const oReal& tol2) const throw()
{
	oReal t = PointHeight(v);
	return (t > _R(0) && (t * t) > (tol2 * n.NormSquared()));
}

oVector2 operator-(const oVector2& l, const oVector2& r) throw()
{
	return oVector2(l.x - r.x, l.y - r.y);
}

void oCreatePlanTransform(const oVector2& min, const oVector2& max, const oReal& y, oMatrix4x4& transform) throw()
{
	transform.e11 = _R(2) / (max.x - min.x);
	transform.e14 = _R(-1) - min.x * transform.e11;
	transform.e23 = _R(2) / (max.y - min.y);
	transform.e24 = _R(-1) - min.y * transform.e23;
	transform.e34 = y;
	transform.e12 = transform.e13
		= transform.e21 = transform.e22
		= transform.e31 = transform.e32 = transform.e33
		= transform.e41 = transform.e42 = transform.e43 = _R(0); 
	transform.e44 = _R(1); 
}

void oCreatePlanPreserveAspect(oVector2& min, oVector2& max, const oReal& aspectRatio) throw()
{
	float tmp = (max.x - min.x) / (max.y - min.y);
	if(tmp > aspectRatio)
	{
		tmp = (max.y - min.y) * ((tmp / aspectRatio) - _R(1)) / _R(2);
		max.y += tmp;
		min.y -= tmp;
	}
	else
	{
		tmp = (max.x - min.x) * ((aspectRatio / tmp) - _R(1)) / _R(2);
		max.x += tmp;
		min.x -= tmp;
	}
}

oPlane3::oPlane3() throw() : n(_R(0), _R(1), _R(0)), d(_R(0)) {}

oPlane3::oPlane3(const oVector3& n_, const oReal& d_) throw() : n(n_), d(d_) {}

oPlane3::oPlane3(const oVector3& a, const oVector3& b, const oVector3& c) throw()
{
	CreateFromPoints(a, b, c);	
}

void oPlane3::CreateFromPoints(const oVector3& a, const oVector3& b, const oVector3& c) throw()
{
	// assuming abc goes clockwise
	n = oCrossProduct(b - a, c - a);
	d = oDotProduct(a, n);
}

oReal oPlane3::PointHeight(const oVector3& v) const throw()
{
	return oDotProduct(v, n) - d;
}

bool oPlane3::IntersectsAABB(const oVector3& centre, const oVector3& halfLengths) const throw()
{
    return oDotProduct(n, centre)
		+ oRealAbs(n.x * halfLengths.x) 
		+ oRealAbs(n.y * halfLengths.y) 
		+ oRealAbs(n.z * halfLengths.z) > d;
}

oFrustrum::oFrustrum(const oVector3& position, const oQuaternion& orientation, const oProjection& projection) throw()
{
	// set the points
	points[0] = position;
	oVector3 tmp(-projection.zFar, projection.zFar, projection.zFar);
	tmp.x /= projection.fovRatio;
	tmp.y /= projection.fovRatio * projection.aspectRatio;
	points[1] = position + oQuaternionRotation(orientation, tmp);
	tmp.x = -tmp.x;
	points[2] = position + oQuaternionRotation(orientation, tmp);
	tmp.y = -tmp.y;
	points[4] = position + oQuaternionRotation(orientation, tmp);
	tmp.x = -tmp.x;
	points[3] = position + oQuaternionRotation(orientation, tmp);
	// set the boundary planes
	bounds[0].CreateFromPoints(points[0], points[1], points[3]); // left
	bounds[1].CreateFromPoints(points[0], points[4], points[2]); // right
	bounds[2].CreateFromPoints(points[1], points[2], points[3]); // back
	bounds[3].CreateFromPoints(points[0], points[2], points[1]); // top
	bounds[4].CreateFromPoints(points[0], points[3], points[4]); // bottom
	bounds[5].n = -bounds[2].n; bounds[5].d = oDotProduct(bounds[5].n, position); // front
}

bool oFrustrum::IntersectsAABB(const oVector3& centre, const oVector3& halfLengths) const throw()
{
	for(ulong i = 0; i < 6; ++i)
	{
		if(!bounds[i].IntersectsAABB(centre, halfLengths))
			return false;
	}
	return true;
}
